let hamiltonian f g =

  let graph_size = Hashtbl.length g in (* le nombre de sommets *)

  (* La fonction auxiliaire faisant vraiment le backtracking *)
  (* v : le sommet que l'on visite
     path : le chemin trouvé jusqu'à présent c'est une liste
            qui contient toujours [v] au sommet.
            Il faut la renverser pour avoir les sommets par
            ordre de parcours
     n:     la longueur du chemin, pour éviter d'appeler à chaque
     fois List.length

  *)

  let rec loop v path n =
    if n = graph_size then (* on a réussit a former un chemin hamiltonien*)
      f (List.rev path)
    else
      let next = Hashtbl.find g v in (* on supose que ça ne peut pas échouer,
                                            un sommet sans voisin est quand meme présent
                                            dans la table avec la liste vide. *)
      List.iter (fun w ->
          (* pour chaque voisin w, s'il n'est pas déjà dans le chemin,
             on essayer de passer par lui.*)
          if not (List.mem w path) then
            loop w (w::path) (n+1)
            (* on utilise des listes, pas besoin de défaire quoi que
               ce soit en revenant en arrière. *)
        ) next
  in
  (* la procédure principale, essaye de trouver un hamiltonien en partant de chaque
     sommet.
     C'est nécessaire dans le cas d'un graphe orienté
  *)
  for v = 0 to graph_size - 1 do
    loop v [v] 1
  done

(* le graphe de test *)
let graph = Hashtbl.create 16


let () =
  Hashtbl.add graph 0 [3];
  Hashtbl.add graph 1 [0;2;4];
  Hashtbl.add graph 2 [7];
  Hashtbl.add graph 3 [4;5];
  Hashtbl.add graph 4 [1;3;6];
  Hashtbl.add graph 5 [6];
  Hashtbl.add graph 6 [4;7];
  Hashtbl.add graph 7 []

(* une fonction d'affichage *)
let print_path l =
  let slist = List.map string_of_int l in
  Printf.printf "%s\n" (String.concat " -> " slist)


let main () = hamiltonian print_path graph

let () = main ()

